-- Bale wrap colorswitcher specialization for FS19
--
-- This mod enables you to switch the bale wrap color at any time, without the need to go to the shop. 
-- Furthermore way more colors are available, and these can be even extended by custom colors.
-- Custom wrap colors can be added in modsSettings/BaleWrapColorSwitcher/wrapColors.xml
--
-- Author: sperrgebiet

WrapColorSwitcher = {}
WrapColorSwitcher.eventName = {}
WrapColorSwitcher.ModName = g_currentModName
WrapColorSwitcher.ModDirectory = g_currentModDirectory
WrapColorSwitcher.Version = "1.0.0.0"

WrapColorSwitcher.debug = fileExists(WrapColorSwitcher.ModDirectory ..'debug')

-- Giants doesn't like prints from mods
--print(string.format('WrapColorSwitcher v%s - DebugMode %s)', WrapColorSwitcher.Version, tostring(WrapColorSwitcher.debug)))

addModEventListener(WrapColorSwitcher)

function WrapColorSwitcher:dp(val, fun, msg) -- debug mode, write to log
	if not WrapColorSwitcher.debug then
		return;
	end

	if msg == nil then
		msg = ' ';
	else
		msg = string.format(' msg = [%s] ', tostring(msg));
	end

	local pre = 'WrapColorSwitcher DEBUG:';

	if type(val) == 'table' then
		--if #val > 0 then
			print(string.format('%s BEGIN Printing table data: (%s)%s(function = [%s()])', pre, tostring(val), msg, tostring(fun)));
			DebugUtil.printTableRecursively(val, '.', 0, 3);
			print(string.format('%s END Printing table data: (%s)%s(function = [%s()])', pre, tostring(val), msg, tostring(fun)));
		--else
		--	print(string.format('%s Table is empty: (%s)%s(function = [%s()])', pre, tostring(val), msg, tostring(fun)));
		--end
	else
		print(string.format('%s [%s]%s(function = [%s()])', pre, tostring(val), msg, tostring(fun)));
	end
end

function WrapColorSwitcher.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(BaleWrapper, specializations)
end

function WrapColorSwitcher:loadMap(name)
	-- Just load the config on the server
	if g_server ~= nil then
		WrapColorSwitcher.userPath = getUserProfileAppPath();
		WrapColorSwitcher.savePath = WrapColorSwitcher.userPath .. 'modsSettings/BaleWrapColorSwitcher/';
		createFolder(WrapColorSwitcher.userPath .. 'modsSettings/');
		createFolder(WrapColorSwitcher.savePath);
		WrapColorSwitcher.xmlFilename = WrapColorSwitcher.savePath .. 'wrapColors.xml';	
		
		WrapColorSwitcher:loadConfig()
	end
end

function WrapColorSwitcher.registerEventListeners(vehicleType)
	local functionNames = {	"onRegisterActionEvents", "onPostLoad", "saveToXMLFile", "onReadStream", "onWriteStream"}

	for _, functionName in ipairs(functionNames) do
		SpecializationUtil.registerEventListener(vehicleType, functionName, WrapColorSwitcher)
	end
end

function WrapColorSwitcher:onRegisterActionEvents(isSelected, isOnActiveVehicle)
	
	local spec = self.spec_wrapColorSwitcher
	spec.actionEvents = {}
	if isSelected then
		local _, actionEventId = self:addActionEvent(spec.actionEvents, 'WrapColorSwitcher_Change', self, WrapColorSwitcher.action_WrapColorSwitcher_Change, false, true, false, true, nil)
		g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL)
        --g_i18n.moneyUnit 1 EURO, 2 Dollar, 3 POUND
        local currencySymbol = "€)"
        if g_i18n.moneyUnit == 2 then
            currencySymbol = "$)"
        elseif g_i18n.moneyUnit == 3 then
            currencySymbol = "£)"
        end

        g_inputBinding:setActionEventText(actionEventId, g_i18n.modEnvironments[WrapColorSwitcher.ModName].texts.WrapColorSwitcher_Change .. currencySymbol)
	end		
end

function WrapColorSwitcher:action_WrapColorSwitcher_Change(actionName, keyStatus, arg3, arg4, arg5)
    if self.spec_baleWrapper.roundBaleWrapper.currentBale == nil and self.spec_baleWrapper.squareBaleWrapper.currentBale == nil then
        local pickerDefaultColor = self.spec_wrapColorSwitcher.colorList[1].color
        if self.spec_wrapColorSwitcher.colorIndex ~= nil and self.spec_wrapColorSwitcher.colorList[self.spec_wrapColorSwitcher.colorIndex] ~= nil then
            pickerDefaultColor = self.spec_wrapColorSwitcher.colorList[self.spec_wrapColorSwitcher.colorIndex].color
        end

        g_gui:showColorPickerDialog({colors = self.spec_wrapColorSwitcher.colorList, defaultColor = pickerDefaultColor, callback = WrapColorSwitcher.colorPickerCallback, target = self, args = self})
    else
        g_currentMission:showBlinkingWarning(g_i18n.modEnvironments[WrapColorSwitcher.ModName].texts.WarningBalerWorking, 5000)
    end
end

function WrapColorSwitcher:colorPickerCallback(selectedIndex, obj)
    if (selectedIndex ~= nil) and (selectedIndex ~= obj.spec_wrapColorSwitcher.colorIndex) then
        WrapColorSwitcher:setCustomColor(obj, selectedIndex, false, false)
    end
end

function WrapColorSwitcher:onPostLoad(savegame)
    
    --we'll get all colors on the wrapper from the server
    if g_server ~= nil then
        if self.spec_wrapColorSwitcher.colorList == nil then
            self.spec_wrapColorSwitcher.colorList = WrapColorSwitcher:getWrapperColorList(self)
        end

        -- The actual onPostLoad ignores the storeItem, so we've to re-set the color based on the stored config index
        if savegame ~= nil then
            local xmlFile = savegame.xmlFile
            local key = savegame.key..".wrapColorSwitcher#colorIndex"
            local colorIndex = Utils.getNoNil(getXMLInt(xmlFile, key), 1)
            self.spec_wrapColorSwitcher.colorIndex = colorIndex
            WrapColorSwitcher:setCustomColor(self, colorIndex, true, false)
        end
    end

    -- To avoid confusing with changing the wrap color in the store, I'll drop all the colors from there, so that the config option disappears
    WrapColorSwitcher:dropWrapperColorList(self)



end

function WrapColorSwitcher:saveToXMLFile(xmlFile, key)
	if self.spec_wrapColorSwitcher ~= nil then
		if self.spec_wrapColorSwitcher.colorIndex ~= nil then
			setXMLInt(xmlFile, key.."#colorIndex", self.spec_wrapColorSwitcher.colorIndex)
		end
	end
end

function WrapColorSwitcher:getDefaultColors()
    local defaultColors = {}

    for i, color in pairs(g_vehicleColors) do
        local newColor = {}
        if color.r ~= nil and color.g ~= nil and color.b ~= nil then
            newColor.name = g_i18n:convertText(color.name)
            newColor.color = {color.r, color.g, color.b, 1}
            newColor.price = 0
            newColor.dailyUpkeep = 0
            newColor.desc = ""
            newColor.isDefault = false
        elseif color.brandColor ~= nil then
            newColor.name = g_i18n:convertText(color.brandColor)
            newColor.color = g_brandColorManager:getBrandColorByName(color.brandColor)
            newColor.price = 0
            newColor.dailyUpkeep = 0
            newColor.desc = ""
            newColor.isDefault = false
        end
        table.insert(defaultColors, newColor)
    end

    return defaultColors
end

function WrapColorSwitcher:getWrapperColorList(obj)
    local storeItem = g_storeManager:getItemByXMLFilename(obj.configFileName)
    local storeColors = storeItem.configurations.wrappingColor

    local colors = {}

    for k, v in pairs(storeColors) do
        table.insert(colors, v)
    end

    for k, v in pairs(WrapColorSwitcher:getDefaultColors()) do
        table.insert(colors, v)
    end

    if WrapColorSwitcher.configColors ~= nil and #WrapColorSwitcher.configColors > 1 then
        for k, v in pairs(WrapColorSwitcher.configColors) do
            table.insert(colors, v)
        end
    end

    return colors
end

function WrapColorSwitcher:dropWrapperColorList(obj)
    local storeItem = g_storeManager:getItemByXMLFilename(obj.configFileName)
    while(#storeItem.configurations.wrappingColor > 1) do
        table.remove(storeItem.configurations.wrappingColor)
    end
end

function WrapColorSwitcher:setCustomColor(obj, selectedIndex, savedItem, noEventSend)
    obj.spec_wrapColorSwitcher.colorIndex = selectedIndex
    WrapColorSwitcher:setVehicleColor(obj, selectedIndex)
    if not savedItem then
        WrapColorSwitcher:deductMoney(obj.ownerFarmId)
    end
    WrapColorSwitcherEvent.sendEvent(obj, selectedIndex, noEventSend)
    

end

function WrapColorSwitcher:deductMoney(farmId)
    -- Just the server can set money events
    if g_server ~= nil then
        local stats = g_farmManager:getFarmById(farmId).stats
        stats:updateStats("expenses", 60)
        g_currentMission:addMoney(-60, farmId, MoneyType.VEHICLE_RUNNING_COSTS, true, true)
    end
end

function WrapColorSwitcher:loadConfig()

	if fileExists(self.xmlFilename) then
		self.saveFile = loadXMLFile('WrapColorSwitcher.loadFile', self.xmlFilename)
		
		if hasXMLProperty(self.saveFile, 'wrapColors') then
            local savedColors = {}
            local i = 0
            while true do
                local key = string.format("wrapColors.color(%d)", i)
                if not hasXMLProperty(self.saveFile, key) then
                    break
                end
                local colorName = getXMLString(self.saveFile, key .. "#name")
                local colorString = getXMLString(self.saveFile, key)

                if colorString ~= nil then
                    local loadedColor = {}
                    loadedColor.name = g_i18n:convertText(colorName)
                    loadedColor.color = ConfigurationUtil.getColorFromString(colorString)
                    loadedColor.price = 0
                    loadedColor.dailyUpkeep = 0
                    loadedColor.desc = ""
                    loadedColor.isDefault = false
                    table.insert(savedColors, loadedColor)
                end
                i = i + 1
            end
			if #savedColors > 0 then
                self.configColors = savedColors
            end
		end
    else
        -- If we've no config yet we'll just create an example one
        self:writeExampleConfig()
        self:loadConfig()
	end
end

function WrapColorSwitcher:writeExampleConfig()
	local configFile = createXMLFile('WrapColorSwitcher.config', self.xmlFilename, "wrapColors")
    if configFile ~= nil then
        -- Just drop some example colors and to make things more 'colorful' :)
        local exampleColors = { "KRONE_GREEN1",
                                "KRONE_YELLOW2",
                                "JOHNDEERE_YELLOW1",
                                "JOHNDEERE_GREEN1",
                                "ROSTSELMASH_YELLOW1",
                                "ROSTSELMASH_ORANGE1",
                                "ROSTSELMASH_GREEN1",
                                "ROSTSELMASH_RED1",
                                "ROSTSELMASH_BROWN1",
                                "ROSTSELMASH_PURPLE1",
                                "DEUTZ_BLUE1",
                                "DEUTZ_BLUE2",
                                "FLIEGL_RED1",
                                "FLIEGL_GREEN1",
                                "STRAUTMANN_GREEN1",
                                "KUHN_RED1",
                                "NEWHOLLAND_YELLOW1",
                                "NEWHOLLAND_BLUE1",
                                "LAMBORGHINI_YELLOW1",
                                "LAMBORGHINI_BLUE1",
                                "LAMBORGHINI_ORANGE1",
                                "LAMBORGHINI_GREEN1",
                                "FENDT_BLACK1",
                                "FENDT_VIOLET1",
                                "FENDT_BRIGHTGREEN",
                                "FENDT_GREY1",
                                "FENDT_BLUE1",
                                "FENDT_RED1",
                                "FENDT_YELLOW1",
                                "FENDT_WHITE1",
								"CLAAS_WHITE1",
								"CLAAS_GREEN1",
								"CLAAS_RED1",
								"CLAAS_DARKGREY1",
                                "KOMATSU_REDMETAL",
                                "PONSSE_BLUE2",
                                "PONSSE_YELLOW2",
                                "JENZ_GREEN1",
                                "SILOKING_BLUE1",
                                "KIROVETS_YELLOW_1",
                                "KIROVETS_RED_1",
                                "KIROVETS_BLUE_1",
                                "STEYR_YELLOW1",
                                "STEYR_GREEN1",
                                "STEYR_BLUE1",
                                "STEYR_ORANGE1",
                                "STEYR_GREY1",
                                "HATZENBICHLER_YELLOW",
                                "HATZENBICHLER_RED",
                                "HATZENBICHLER_GREEN",
                                "HATZENBICHLER_DARKBLUE1",
                                "LINDNER_BEIGE1",
                                "LINDNER_BLUE1",
                                "LINDNER_GREEN1",
                                "LINDNER_RED1"
                                }

        setXMLString(configFile, "wrapColors.color(0)#name", "Kommunal/Municipal")
        setXMLString(configFile, "wrapColors.color(0)", "0.8879 0.1878 0.0037 1")
        setXMLString(configFile, "wrapColors.color(1)#name", "RAL 1006 - Maisgelb/Maize yellow")
        setXMLString(configFile, "wrapColors.color(1)", "0.8862745 0.5647059 0 1")
        setXMLString(configFile, "wrapColors.color(2)#name", "RAL 1021 - Rapsgelb/Colza yellow")
        setXMLString(configFile, "wrapColors.color(2)", "0.9647059 0.7137255 0 1")
        setXMLString(configFile, "wrapColors.color(3)#name", "RAL 1026 - Leuchtgelb/Luminous yellow")
        setXMLString(configFile, "wrapColors.color(3)", "1 1 0 1")
        setXMLString(configFile, "wrapColors.color(4)#name", "RAL 2003 - Pastellorange/Pastel orange")
        setXMLString(configFile, "wrapColors.color(4)", "0.9647059 0.4705882 0.1568628 1")
        setXMLString(configFile, "wrapColors.color(5)#name", "RAL 4003 - Erikaviolett/Heather violet")
        setXMLString(configFile, "wrapColors.color(5)", "0.7686275 0.3803922 0.5490196 1")
        setXMLString(configFile, "wrapColors.color(6)#name", "RAL 4008 - Signalviolett/Signal violet")
        setXMLString(configFile, "wrapColors.color(6)", "0.5176471 0.2980392 0.5098040 1")
        setXMLString(configFile, "wrapColors.color(7)#name", "RAL 5012 - Lichtblau/Light blue")
        setXMLString(configFile, "wrapColors.color(7)", "0 0.5372549 0.7137255 1")

        local i = 8
        for _, v in pairs(exampleColors) do
            setXMLString(configFile, string.format("wrapColors.color(%d)#name", i), v)
            setXMLString(configFile, string.format("wrapColors.color(%d)", i), v)
            i = i + 1
        end

        saveXMLFile(configFile)
    end
end

function WrapColorSwitcher:setVehicleColor(object, configColorId)
    local xmlFile = object.xmlFile
    local color
    if object.spec_wrapColorSwitcher.colorIndex ~= nil and object.spec_wrapColorSwitcher.colorList[object.spec_wrapColorSwitcher.colorIndex] ~= nil then
        color = object.spec_wrapColorSwitcher.colorList[object.spec_wrapColorSwitcher.colorIndex].color
    else
        color = ConfigurationUtil.getColorByConfigId(object, "wrappingColor", 1)
    end
    if color ~= nil then
        local r,g,b,mat = unpack(color);
        local i = 0;
        while true do
            local colorKey = string.format("vehicle.%sConfigurations.colorNode(%d)", "wrappingColor", i)
            if not hasXMLProperty(xmlFile, colorKey) then
                break;
            end
            local node = I3DUtil.indexToObject(object.components, getXMLString(xmlFile, colorKey .. "#node"), object.i3dMappings);
            if node ~= nil then
                if getHasClassId(node, ClassIds.SHAPE) then
                    if mat == nil then
                        _,_,_,mat = getShaderParameter(node, "colorScale")
                    end
                    if Utils.getNoNil(getXMLBool(xmlFile, colorKey .. "#recursive"), false) then
                        I3DUtil.setShaderParameterRec(node, "colorScale", r, g, b, mat);
                    else
                        setShaderParameter(node, "colorScale", r, g, b, mat, false);
                    end
                else
                    print("Warning: Could not set vehicle color to '"..getName(node).."' because node is not a shape!")
                end
            end
            i = i + 1;
        end
    end
end

function WrapColorSwitcher:getStringFromColor(colorTable)
    local colorString = ""
    if #colorTable == 4 then
        colorString = string.format("%f %f %f %f", colorTable[1], colorTable[2], colorTable[3], colorTable[4])
    else
        colorString = "0 0 0 0"
    end
    return colorString
end

function WrapColorSwitcher:updateWrappingState(superFunc, t, noEventSend)
    local spec = self.spec_baleWrapper
    t = math.min(t, 1)
    local wrapperRot = 0
    if spec.currentWrapper.animCurve ~= nil then
        local v = spec.currentWrapper.animCurve:get(t)
        if v ~= nil then
            setRotation(spec.currentWrapper.baleNode, v[1]%(math.pi*2), v[2]%(math.pi*2), v[3]%(math.pi*2))
            setRotation(spec.currentWrapper.wrapperNode, v[4]%(math.pi*2), v[5]%(math.pi*2), v[6]%(math.pi*2))
            wrapperRot = v[3 + spec.currentWrapper.wrapperRotAxis]
        elseif spec.currentWrapper.animations["wrapBale"].animName ~= nil then
            t = self:getAnimationTime(spec.currentWrapper.animations["wrapBale"].animName)
        end
        if spec.currentWrapper.currentBale ~= nil then
            local bale = NetworkUtil.getObject(spec.currentWrapper.currentBale)
            if bale ~= nil then
                if not self:getIsBaleFillTypeSkiped(bale) then
                    if self.isServer then
                        local wrappingState = t
                        if table.getn(spec.currentWrapper.wrappingStateCurve.keyframes) > 0 then
                            wrappingState = spec.currentWrapper.wrappingStateCurve:get(t)
                        end
                        bale:setWrappingState(wrappingState)
                    end
                    if bale.setColor ~= nil then
                        local color
                        if self.spec_wrapColorSwitcher.colorIndex ~= nil then
                            color = self.spec_wrapColorSwitcher.colorList[self.spec_wrapColorSwitcher.colorIndex].color
                        else
                            color = ConfigurationUtil.getColorByConfigId(self, "wrappingColor")
                        end
                        if color ~= nil then
                            local r, g, b, a = unpack(color)
                            bale:setColor(r, g, b, a)
                        end
                    end
                end
            end
        end
    end
    self:updateWrapNodes(t > 0, false, t, wrapperRot)
    if t == 1 then
        if self.isServer and spec.baleWrapperState == BaleWrapper.STATE_WRAPPER_WRAPPING_BALE and not noEventSend then
            g_server:broadcastEvent(BaleWrapperStateEvent:new(self, BaleWrapper.CHANGE_WRAPPING_BALE_FINSIHED), true, nil, self)
        end
    end
end

-- Not super efficient to sync all colors (also global ones), but it was the easiest way to do it
function WrapColorSwitcher:onReadStream(streamId, connection)
    local spec = self.spec_wrapColorSwitcher
    spec.colorIndex = streamReadInt8(streamId)
    local numColorList = streamReadInt8(streamId)
    local i = 1
    local colorList = {}
    while(i <= numColorList) do
        local loadedColor = {}
        loadedColor.name = streamReadString(streamId)
        loadedColor.color = ConfigurationUtil.getColorFromString(streamReadString(streamId))
        loadedColor.price = 0
        loadedColor.dailyUpkeep = 0
        loadedColor.desc = ""
        loadedColor.isDefault = false
        table.insert(colorList, loadedColor)
        i = i + 1
    end
    if #colorList > 0 then
        spec.colorList = colorList
    end
	WrapColorSwitcher:setCustomColor(self, self.spec_wrapColorSwitcher.colorIndex, true, true)
end

function WrapColorSwitcher:onWriteStream(streamId, connection)
    local spec = self.spec_wrapColorSwitcher
    streamWriteInt8(streamId, spec.colorIndex)
    streamWriteInt8(streamId, #spec.colorList)
    local i = 1
    while(i <= #spec.colorList) do
        streamWriteString(streamId, spec.colorList[i].name)
        streamWriteString(streamId, WrapColorSwitcher:getStringFromColor(spec.colorList[i].color))
        i = i + 1
    end
end

--
---- MP Events
--
WrapColorSwitcherEvent = {}
WrapColorSwitcherEvent_mt = Class(WrapColorSwitcherEvent, Event)

InitEventClass(WrapColorSwitcherEvent, "WrapColorSwitcherEvent")

function WrapColorSwitcherEvent:emptyNew()
    local self = Event:new(WrapColorSwitcherEvent_mt)
	self.className = "WrapColorSwitcherEvent"
    return self
end

function WrapColorSwitcherEvent:new(object, colorIndex)
    local self = WrapColorSwitcherEvent:emptyNew()
    self.object = object
	self.colorIndex = colorIndex
    return self
end

function WrapColorSwitcherEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
	self.colorIndex  = streamReadInt8(streamId)
	self:run(connection)
end

function WrapColorSwitcherEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteInt8(streamId, self.colorIndex)
end

function WrapColorSwitcherEvent:run(connection)
	WrapColorSwitcher:setCustomColor(self.object, self.colorIndex, false, true)
    if not connection:getIsServer() then
        g_server:broadcastEvent(WrapColorSwitcherEvent:new(self.object, self.colorIndex), nil, connection, self.object)
    end
end

function WrapColorSwitcherEvent.sendEvent(vehicle, colorIndex, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(WrapColorSwitcherEvent:new(vehicle, colorIndex), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(WrapColorSwitcherEvent:new(vehicle, colorIndex))
		end
	end
end

-- We've to overwrite the default Giants function, as it just takes colors which are defined as Configuration for a bale wrapper
BaleWrapper.updateWrappingState = Utils.overwrittenFunction(BaleWrapper.updateWrappingState, WrapColorSwitcher.updateWrappingState)